# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# ======================================================================

__version__= '1.1'

from Globals import HTML, HTMLFile, DTMLFile, MessageDialog, Persistent
from ImageFile import ImageFile  
import OFS,Acquisition, AccessControl.Role
import Globals, sys, traceback
from OFS.DTMLDocument import DTMLDocument
from AccessControl import getSecurityManager


from OFS.Document import Document

import os,stat,time,string,Products.PythonScripts
from Products.ZSQLMethods.SQL import SQL

from OFS.History import Historical, html_diff
from OFS.SimpleItem import SimpleItem
from OFS.Cache import Cacheable



ProductName = 'Folder Indexer'
ProductFolder = 'FolderIndexer'




class zProductDocument(OFS.SimpleItem.Item,DTMLDocument):
    """Zope Product object class"""    
       
    meta_type = ProductName
    icon = 'misc_/%s/icon' % (ProductFolder)

 
    #manage_about = HTMLFile('README', globals())
    manage_editfields = HTMLFile('dtml/edit_fields', globals())
    manage_methods = HTMLFile('dtml/methods', globals())
    manage_utils = HTMLFile('dtml/utils', globals())
    manage_opt = HTMLFile('dtml/options', globals())
    datasel = HTMLFile('dtml/datasel', globals())
    errorinternalmethod = HTMLFile('dtml/error', globals())
    
    def index_html(self):
        """ default html file returned when view object """
        return "nothing here"


    #manage tabs
    manage_options=(
                    {'label':'Utils', 'action':'manage_utils','help':(ProductFolder,'utils.stx')},
                    {'label':'Options','action':'manage_opt','help':(ProductFolder,'options.stx')},                    
                    {'label':'Methods', 'action':'manage_methods','help':(ProductFolder,'methods.stx')},
                    #{'label':'Description', 'action':'manage_main'},
                    #{'label':'Properties', 'action':'manage_propertiesForm'},
                    #{'label':'View', 'action':''},
                    #{'label':'Proxy', 'action':'manage_proxyForm', 'help': ('OFSP','DTML-DocumentOrMethod_Proxy.stx')},
                    
                   ) + Historical.manage_options + SimpleItem.manage_options + Cacheable.manage_options
                   

    sql_server_type_dict={
        1:'Generic SQL Server',
        2:'Microsoft SQL Server',
        3:'Oracle SQL Server',
        4:'Postgres SQL Server',
        5:'MySQL SQL Server'
        }

    def __init__(self, id, title, link_base, htdocs_dir, batch_size, delay_batch, database):
        """ init method """
        self.ProductFolder=ProductFolder
        self.id=id
        self.title=title
        self.link_base=link_base
        self.htdocs_dir=htdocs_dir
        self.batch_size=batch_size
        self.delay_batch=delay_batch
        self.debug=0
        self.globals=''
        self._vars=''
        self.database=database
        self.sql_server_type=1
        self.sz_filepath=255
        self.sz_title=255
        self.sz_word=128
        self.sq_table_words=str(id)+'_words'
        self.sq_table_files=str(id)+'_files'
        #create default values
        self.create_internal_methods()
        self.raw="""
Type your comments about this form here.
Type description of this form here.

Here is the body of this document.
"""


    def update_words(self,REQUEST=None):
        """ Update words - test if file is changed or missing
        from table and index it,
        stop after self.batch_size files!!!!
        wait self.delay_batch seconds then reload html page ...
        """
        self.cnt=self.batch_size
        if self.debug==1: print "Start indexing"
        self.file_search(self.htdocs_dir[:-1])
        if self.debug==1: print "End indexing"
        return """
<html>
<meta http-equiv="REFRESH" content="%d">
<body>
<h1>Ok!</h1>Indexed %d files!<br>Wait auto refresh for next %d files!
""" % (self.delay_batch,self.batch_size,self.batch_size)
        



    def update_files(self,REQUEST=None):
        """ Update files - delete file from table if no longer exist"""
        self.SQL_UpdateFiles()
        return """
<html>
<body>
<h1>Ok!</h1>Updated files table!
"""


    def manage_sql(self,act='',REQUEST=None):
        """ Create/Drop tables and seq """
        sql_server_type = int(self.sql_server_type)
        self.create_internal_methods()
        if act == 'Create table files':
            try:
                if sql_server_type in [3,4]: self.SQLCreateSEQfiles() #only for postgres and oracle
                self.SQLCreatefiles()
                return Globals.MessageDialog( title='Create', message="<p align='left'>Ok! Table <b>%s</b> created.<br><br>See generated SQL method <a href='SQLCreatefiles/manage_main'>here</a>!</p>" % (self.sq_table_files), action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Create', message="<p align='left'><font color='red'>Error! Can't create table.<br>See generated SQL method <a href='SQLCreatefiles/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)
        if act == 'Create table words':
            try:
                #if sql_server_type in [3,4]: self.SQLCreateSEQwords() #only for postgres and oracle
                self.SQLCreatewords()
                return Globals.MessageDialog( title='Create', message="<p align='left'>Ok! Table <b>%s</b> created.<br><br>See generated SQL method <a href='SQLCreatewords/manage_main'>here</a>!</p>" % (self.sq_table_words), action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Create', message="<p align='left'><font color='red'>Error! Can't create table.<br>See generated SQL method <a href='SQLCreatewords/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)
        if act == 'Drop table files':
            try:
                if sql_server_type in [3,4]: self.SQLDropSEQfiles() #only for postgres and oracle
                self.SQLDropfiles()
                return Globals.MessageDialog( title='Drop', message="<p align='left'>Ok! Table erased (dropped).<br><br>See generated SQL method <a href='SQLDropfiles/manage_main'>here</a>!</p>", action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Drop', message="<p align='left'><font color='red'>Error! Table may not exist.<br>Or somebody locked it.<br><br>See generated SQL method <a href='SQLDropfiles/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)
        if act == 'Drop table words':
	    print "aaaaa"
            try:
                self.SQLDropwords()
                return Globals.MessageDialog( title='Drop', message="<p align='left'>Ok! Table erased (dropped).<br><br>See generated SQL method <a href='SQLDropwords/manage_main'>here</a>!</p>", action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Drop', message="<p align='left'><font color='red'>Error! Table may not exist.<br>Or somebody locked it.<br><br>See generated SQL method <a href='SQLDropwords/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)
        if REQUEST is not None:
            return Globals.MessageDialog( title='Edited', message='<strong>%s</strong> has been edited.' % self.id, action ='manage_opt',)


    def manage_opt_action(self, title, link_base='', htdocs_dir='', batch_size=10, delay_batch=30, database='', sql_server_type=0, debug='',REQUEST=None):
        """ update properties  """
        self.title = title
        self.link_base=link_base
        self.htdocs_dir=htdocs_dir.replace('\\','/')
        try:    self.batch_size=int(batch_size)
        except: self.batch_size=10
        try:    self.delay_batch=int(delay_batch)
        except: self.delay_batch=30
        if self.link_base=='': self.link_base='http://server/folder'
        if self.link_base[-1]!='/':
            self.link_base=self.link_base+'/'
        if self.htdocs_dir=='': self.htdocs_dir='/webarea'
        if self.htdocs_dir[-1]!='/':
            self.htdocs_dir=self.htdocs_dir+'/'
        self.database = database
        self.sql_server_type = sql_server_type
        sql_server_type = int(self.sql_server_type)
        #set SQL field size according with SQL server type
        if sql_server_type == 1: #Generic SQL
            self.sz_filepath=255
            self.sz_title=255
            self.sz_word=128
        if sql_server_type == 2: #Microsoft
            self.sz_filepath=2048
            self.sz_title=512
            self.sz_word=128
        if sql_server_type == 3: #Oracle
            self.sz_filepath=2048
            self.sz_title=512
            self.sz_word=128
        if sql_server_type == 4: #Postgres
            self.sz_filepath=2048
            self.sz_title=512
            self.sz_word=128
        if sql_server_type == 5: #mySQL
	    #mySQL allow max 255 chars as fieldsize
            self.sz_filepath=255
            self.sz_title=255
            self.sz_word=128
        
        if debug=='': self.debug=0
        else: self.debug=1
        #update SQL methods
        self.create_internal_methods()
        if REQUEST is not None:
            return Globals.MessageDialog( title='Edited', message='<strong>%s</strong> has been edited.' % self.id, action ='manage_opt',)



    def create_internal_methods(self):
        
        #Create python script search method
        self.search = Products.PythonScripts.PythonScript.PythonScript('search')
        self.search.ZPythonScript_setTitle('Search python script')
        self.search.ZPythonScript_edit('','''
REQUEST = context.REQUEST
RESPONSE =  REQUEST.RESPONSE
SQL1 = context.SQL1
SQL2 = context.SQL2

try:
  #try to get query search expression
  qor = REQUEST.qor
except:
  qor=''

#get radio buttons status
try:    any = REQUEST.any
except: any = 'yes'

#get statistics flag status
try:    statistics = REQUEST.statistics
except: statistics = ''

#get modification time flag status
try:    modtime = REQUEST.modtime
except: modtime = ''

if any=='yes':
  xx1=' CHECKED'
  xx2=''
else:
  any='no'
  xx1=''
  xx2=' CHECKED'

if statistics=='yes': xx=' CHECKED'
else: xx=''
if modtime=='yes': yy=' CHECKED'
else: yy=''

print """
<html>
<head>
 <title>Site Search</title>
</head>
<body>
<h1>Search</h1>
<form method="get">
Search: <input type="text" name="qor" value="%%s"><br>
<input type="radio" name="any" value="yes"%%s>Match any words<br>
<input type="radio" name="any" value="no"%%s>Match all words<br>
<input type="checkbox" name="statistics" value="yes"%%s>Include statistics<br>
<input type="checkbox" name="modtime" value="yes"%%s>Include last modification time<br>
<input type="submit" value="Search">
</form>
""" %% (qor,xx1,xx2,xx,yy)

L=[] #list of searched words
if qor!='':
  for i in qor.split(" "):
    if i!="": L.append(i)

F=[] #file id list
  
if L!=[] and any=='yes':
  q="select distinct fileid from %s where "
  for i in range(len(L)):
    if i>0: q=q+" or"
    q=q+" word='%%s' " %% (L[i])
  #print q,"<br>"
  for i in SQL1(q=q): F.append(i.fileid)

if L!=[] and any=='no':
  #get files for L[0]
  for i in SQL1(q="select fileid from %s where word='%%s'" %% (L[0])):
   F.append(i.fileid)
  #Erase items from F if that file do not contain the word from var w
  for w in L[1:]:
    tmp_F=[]
    for id in F:
      fid=0
      q="select fileid from %s where fileid=%%d and word='%%s'" %% (id,w)
      for i in SQL1(q=q):
        fid=i.fileid
      if fid!=0: tmp_F.append(id)
    F=tmp_F


if F==[]: print "<b>Nothing found!</b>"
else:
  if statistics=='yes': print "<b>Display statistics of found words.</b><br>"
  print "<b>Found %%d files:</b><br>" %% (len(F))
  for id in F:
   for i in SQL1(q="select title,filepath,modtime from %s where id=%%d" %% (id)):
      print '<img src="/misc_/OFSP/dtmldoc.gif"> <a href="%s%%s">%%s</a>' %% (i.filepath[%d:],i.title)
      if modtime=='yes':
        print " (last update : <b>%%s</b>)" %% (context.ZopeTime(int(i.modtime)).strftime('%%A, %%B %%d, %%Y'))
      print "<br>"
      if statistics=='yes':
        print "&nbsp;"*10
        for w in L:
          q="select count from %s where word='%%s' and fileid=%%d" %% (w,id)
          for j in SQL2(q=q):
            print "<b>",w,"</b>:",j.count,"&nbsp;"*3
        print "<br>"
  
print "</body></html>"

return printed
''' % (self.sq_table_words,self.sq_table_words,self.sq_table_words,self.sq_table_files,self.link_base,len(self.htdocs_dir),self.sq_table_words,) )

        start_val = 1 #for files table this is start id value
        
        #Create generic SQL1
        self.SQL1=SQL('SQL1', 'Generic SQL Method #1', self.database, 'q', '''
<dtml-comment>
  This is generic SQL method #1.
  Is used to be called from other methods.
  DO NOT EDIT THIS METHOD!
  Create it in your applications or put it in your Zope root to be available to all apps
</dtml-comment>

<dtml-var q>
''')

        #Create generic SQL2 
        self.SQL2=SQL('SQL2', 'Generic SQL Method #2', self.database, 'q', '''
<dtml-comment>
  This is generic SQL method #2.
  Is used to be called from other methods.
  DO NOT EDIT THIS METHOD!
  Create it in your applications or put it in your Zope root to be available to all apps
</dtml-comment>

<dtml-var q>
''')

        #SQL create sequence (files)
        sql_server_type = int(self.sql_server_type)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        if sql_server_type == 3: #Oracle
            q = q + 'create sequence %s_seq_id start with %d\n' % (self.sq_table_files,start_val)
        if sql_server_type == 4: #Postgres
            q = q + 'create sequence %s_seq_id start %d\n' % (self.sq_table_files,start_val)
        self.SQLCreateSEQfiles=SQL('SQLCreateSEQfiles', 'Create sequence - files', self.database, '', q)
        
        #SQL create table (files)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        q = q + '''
create table '''+str(self.sq_table_files)+'\n(\n id integer'
        #if sql_server_type == 1 -> Generic SQL Server                        
        if sql_server_type == 2: #Microsoft SQL
            q=q+' IDENTITY (%d, 1)' % (start_val)
        if sql_server_type == 3: #Oracle - default value NOT work so I'll use seq.nextval in insert statement
            pass
        if sql_server_type == 4: #Postgres
            q=q+" default nextval('%s_seq_id')" % (self.sq_table_files)
        if sql_server_type == 5: #MySQL
            q=q+' auto_increment default %d' % (start_val)
        q=q+' PRIMARY KEY,\n filepath varchar(%d),\n modtime integer,\n title varchar(%d)\n)' % (self.sz_filepath,self.sz_title)
        self.SQLCreatefiles=SQL('SQLCreatefiles', 'Create Table files', self.database, '', q)


        #SQL drop sequence (files)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        if sql_server_type == 3: #Oracle
            q = q + 'drop sequence %s_seq_id\n' % (self.sq_table_files)
        if sql_server_type == 4: #Postgres
            q = q + 'drop sequence %s_seq_id\n' % (self.sq_table_files)
        self.SQLDropSEQfiles=SQL('SQLDropSEQfiles', 'Drop sequence', self.database, '', q)

        #SQL drop table (files)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        q = q + '''drop table '''+str(self.sq_table_files)
        self.SQLDropfiles=SQL('SQLDropfiles', 'Drop Table', self.database, '', q)



        #SQL drop table (words)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        q = q + '''drop table '''+str(self.sq_table_words)
        self.SQLDropwords=SQL('SQLDropwords', 'Drop Table', self.database, '', q)





        #SQL create table (words)
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        q = q + '''
create table '''+str(self.sq_table_words)+'\n(\n'
        q=q+' word varchar(%d),\n fileid integer\n, count integer\n)' % (self.sz_word)
        self.SQLCreatewords=SQL('SQLCreatewords', 'Create Table words', self.database, '', q)


    def file_search(self,dir):
        for f in os.listdir(dir):
            if self.cnt<0 : break
            pathname = '%s/%s' % (dir,f)
            filestat = os.stat(pathname)
            mode = filestat[stat.ST_MODE]
            size = filestat[stat.ST_SIZE]
            if stat.S_ISDIR(mode):
                #print "dir",pathname
                if f[0] != '_':
                  #skip private frontpage folders
                  self.file_search(pathname)
            else:
                if pathname[-3:] in ['htm','asp','txt'] or pathname[-4:]=='html':
                    if size < 1024*150:
                      self.SQL_AddSearchedFile(pathname,filestat[stat.ST_MTIME])
                    else:
                      print "File too big, not indexed; size=",size/1024,"k\nfile=",pathname


    def GetTitle(self,pathname):
      ix=1+pathname.rfind("/")
      title=pathname[ix:]
      f=open(pathname,"rt")
      doc=f.read()
      f.close()
      ix1=doc.find("<title>")
      if ix1==-1: ix1=doc.find("<TITLE>")
      if ix1==-1: return title.replace("'","`")
      ix2=doc.find("</title>")
      if ix2==-1: ix2=doc.find("</TITLE>")
      if ix2==-1: return title.replace("'","`")
      title=doc[ix1+7:ix2]
      if title=='': title="(untitled)"
      title=title.replace("'","`")
      return title



    def SQL_AddSearchedFile(self,filename,mtime):
      flag=0
      try:
        #already added
        test=0
        for i in self.SQL1(q="select id,modtime,title from %s where filepath='%s'" % (self.sq_table_files,filename)):
          fileid=i.id
          test=1
          modtime=i.modtime
          title=i.title
          if mtime!=modtime:
              self.cnt=int(self.cnt)-1
              if self.cnt<0 : return
              #file modified
              if self.debug==1: print "update",self.cnt,filename
              title=self.GetTitle(filename)
              self.SQL1(q="update %s set modtime=%d, title='%s' where id=%d" % (self.sq_table_files,mtime,title,fileid))
              #delete all its words
              self.SQL_EraseWords(fileid)
              flag=1
        if test==0:
            raise Exception("Force exception")
              
      except:
        #print "Error:"
        #type, val, tb = sys.exc_info()
        #sys.stderr.write(string.join(traceback.format_exception(type, val, tb), ''))
        #del type, val, tb

        #add it
        self.cnt=int(self.cnt)-1
        if self.cnt<0 : return
        if self.debug==1: print "add",self.cnt,filename
        title=self.GetTitle(filename)
        sql_server_type = int(self.sql_server_type)
        if sql_server_type==1: #generic SQL server
            #we must generate an id
            q="select max(id) as maxid from %s" % (self.sq_table_files)
            maxid=0
            for i in self.SQL1(q=q):
                maxid = str(i.maxid)
            if str(maxid) in ['','None']:
                maxid=0
            try: maxid=int(maxid)+1
            except:
                try: maxid=int(float(maxid))+1 #oracle return int as float sometimes ...
                except: maxid=1
            q="insert into %s(id,filepath,modtime,title) values(%d,'%s',%d,'%s')" % (self.sq_table_files,maxid,filename,mtime,title)
        elif sql_server_type==3: #Oracle
            q="insert into %s(id,filepath,modtime,title) values(%s_seq_id.nextval,'%s',%d,'%s')" % (self.sq_table_files,self.sq_table_files,filename,mtime,title)
        else:
            #is not generic SQL and not Oracle which menas it have an autoid with default value
            q="insert into %s(filepath,modtime,title) values('%s',%d,'%s')" % (self.sq_table_files,filename,mtime,title)
        self.SQL1(q=q)
        for i in self.SQL2(q="select id from %s where filepath='%s'" % (self.sq_table_files,filename)):
          fileid=i.id
          flag=1
      if flag==1: self.SQL_IndexWords(filename,fileid)



    def SQL_EraseWords(self,fileid):
      try:
        self.SQL1(q="delete from %s where fileid=%d" % (self.sq_table_words,fileid))
        #print "Deleted",fileid
      except:
        if self.debug==1: print "error delete",fileid

    def SQL_IndexWords(self,filename,fileid):
      #print "Index",filename
      file = open(filename,'rt')
      body = file.read()
      file.close()
      #body=body.lower()
      body=body.replace("\t"," ")
      body=body.replace("\n"," ")
      body=body.replace("\r"," ")
      body=body.replace("."," ")
      body=body.replace(","," ")
      body=body.replace(":"," ")
      body=body.replace(";"," ")
      #erase all <......>
      while 1:
        ix1=ix2=0
        ix1=body.find("<")
        if ix1==-1: break
        ix2=body.find(">",ix1)
        if ix2==-1: break
        body=body[:ix1]+" "+body[ix2+1:]

      #L=[]
      for i in body.split(" "):
        if len(i)>2 and self.only_chars(i):
            #L.append(i)
            q="select count from %s where word='%s' and fileid=%s" % (self.sq_table_words,i,fileid)
            count=0
            for j in self.SQL1(q=q):
                count=int(j.count)+1
            if count!=0:
                q="update %s set count=%d where word='%s' and fileid=%d" % (self.sq_table_words,count,i,fileid)
            else:
                q="insert into %s(word,fileid,count) values('%s',%d,1)" % (self.sq_table_words,i,fileid)
            #print q
            try: self.SQL1(q=q)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                if self.debug==1: print errmsg
            #print L




    def only_chars(self,s):
      flag=0
      for i in s:
        if i not in string.letters:
          flag=1
          break
      return not flag



    def SQL_UpdateFiles(self):
      SQL1=self.SQL1
      SQL2=self.SQL2
      if self.debug==1: print "Update files"
      for i in SQL1(q='select filepath,id from %s' % (self.sq_table_files,)):
        if os.access(i.filepath,os.F_OK) == 0:
          #no file
          SQL2(q="delete from %s where id=%d" % (self.sq_table_files,i.id))
          self.SQL_EraseWords(i.id)
          if self.debug==1: print "no",i.filepath,'-SQL deleted!'
      return "Ok!"


Globals.default__class_init__(zProductDocument)


